@@ -54,7 +54,8 @@ namespace
         CiftiXML m_xml;//because we need to parse it to set up the dimensions anyway
     public:
         CiftiOnDiskImpl(const AString& filename);//read-only
-        CiftiOnDiskImpl(const AString& filename, const CiftiXML& xml, const CiftiVersion& version, const bool& swapEndian);//make new empty file with read/write
+        CiftiOnDiskImpl(const AString& filename, const CiftiXML& xml, const CiftiVersion& version, const bool& swapEndian,
+                        const int16_t& datatype, const bool& rescale, const double& minval, const double& maxval);//make new empty file with read/write
         void getRow(float* dataOut, const std::vector<int64_t>& indexSelect, const bool& tolerateShortRead) const;
         void getColumn(float* dataOut, const int64_t& index) const;
         const CiftiXML& getCiftiXML() const { return m_xml; }
@@ -62,6 +63,7 @@ namespace
         bool isSwapped() const { return m_nifti.getHeader().isSwapped(); }
         void setRow(const float* dataIn, const std::vector<int64_t>& indexSelect);
         void setColumn(const float* dataIn, const int64_t& index);
+        void close();
     };
     
     class CiftiMemoryImpl : public CiftiFile::WriteImplInterface
@@ -135,17 +137,22 @@ CiftiFile::WriteImplInterface::~WriteImplInterface()
 {
 }
 
+CiftiFile::CiftiFile()
+{
+    m_endianPref = NATIVE;
+    setWritingDataTypeNoScaling();//default argument is float32
+}
+
 CiftiFile::CiftiFile(const AString& fileName)
 {
     m_endianPref = NATIVE;
+    setWritingDataTypeNoScaling();//default argument is float32
     openFile(fileName);
 }
 
 void CiftiFile::openFile(const AString& fileName)
 {
-    m_writingImpl.reset();
-    m_readingImpl.reset();//to make sure it closes everything first, even if the open throws
-    m_dims.clear();
+    close();//to make sure it closes everything first, even if the open throws
     boost::shared_ptr<CiftiOnDiskImpl> newRead(new CiftiOnDiskImpl(pathToAbsolute(fileName)));//this constructor opens existing file read-only
     m_readingImpl = newRead;//it should be noted that if the constructor throws (if the file isn't readable), new guarantees the memory allocated for the object will be freed
     m_xml = newRead->getCiftiXML();
@@ -161,15 +168,33 @@ void CiftiFile::setWritingFile(const AString& fileName, const CiftiVersion& writ
     m_endianPref = endian;
 }
 
+void CiftiFile::setWritingDataTypeNoScaling(const int16_t& type)
+{
+    m_writingDataType = type;//could do some validation here
+    m_doWriteScaling = false;
+    m_minScalingVal = -1.0;//these scaling values should never be used, but don't leave them uninitialized
+    m_maxScalingVal = 1.0;
+    m_writingImpl.reset();//prevent writing to previous writing implementation, let the next set...() set up for writing
+}
+
+void CiftiFile::setWritingDataTypeAndScaling(const int16_t& type, const double& minval, const double& maxval)
+{
+    m_writingDataType = type;//could do some validation here
+    m_doWriteScaling = true;
+    m_minScalingVal = minval;
+    m_maxScalingVal = maxval;
+    m_writingImpl.reset();//prevent writing to previous writing implementation, let the next set...() set up for writing
+}
+
 void CiftiFile::writeFile(const AString& fileName, const CiftiVersion& writingVersion, const ENDIAN& endian)
 {
     if (m_readingImpl == NULL || m_dims.empty()) throw CiftiException("writeFile called on uninitialized CiftiFile");
     bool writeSwapped = shouldSwap(endian);
-    AString canonicalFilename = pathToCanonical(fileName);//NOTE: returns EMPTY STRING for nonexistant file
+    AString canonicalFilename = pathToCanonical(fileName);//NOTE: returns EMPTY STRING for nonexistent file
     const CiftiOnDiskImpl* testImpl = dynamic_cast<CiftiOnDiskImpl*>(m_readingImpl.get());
     bool collision = false, hadWriter = (m_writingImpl != NULL);
     if (testImpl != NULL && canonicalFilename != "" && pathToCanonical(testImpl->getFilename()) == canonicalFilename)
-    {//empty string test is so that we don't say collision if both are nonexistant - could happen if file is removed/unlinked while reading on some filesystems
+    {//empty string test is so that we don't say collision if both are nonexistent - could happen if file is removed/unlinked while reading on some filesystems
         if (m_onDiskVersion == writingVersion && (dontRewrite(endian) || writeSwapped == testImpl->isSwapped())) return;//don't need to copy to itself
         collision = true;//we need to copy to memory temporarily
         boost::shared_ptr<WriteImplInterface> tempMemory(new CiftiMemoryImpl(m_xml));//because tempRead is a ReadImpl, can't be used to copy to
@@ -177,7 +202,8 @@ void CiftiFile::writeFile(const AString& fileName, const CiftiVersion& writingVe
         m_readingImpl = tempMemory;//we are about to make the old reading impl very unhappy, replace it so that if we get an error while writing, we hang onto the memory version
         m_writingImpl.reset();//and make it re-magic the writing implementation again if it tries to write again
     }
-    boost::shared_ptr<WriteImplInterface> tempWrite(new CiftiOnDiskImpl(pathToAbsolute(fileName), m_xml, writingVersion, writeSwapped));
+    boost::shared_ptr<WriteImplInterface> tempWrite(new CiftiOnDiskImpl(pathToAbsolute(fileName), m_xml, writingVersion, writeSwapped,
+                                                                        m_writingDataType, m_doWriteScaling, m_minScalingVal, m_maxScalingVal));
     copyImplData(m_readingImpl.get(), tempWrite.get(), m_dims);
     if (collision)//if we rewrote the file, we need the handle to the new file, and to dump the temporary in-memory version
     {
@@ -190,6 +216,22 @@ void CiftiFile::writeFile(const AString& fileName, const CiftiVersion& writingVe
     }
 }
 
+void CiftiFile::close()
+{
+    if (m_writingImpl != NULL)
+    {
+        m_writingImpl->close();//only writing implementations should ever throw errors on close, and specifically only on-disk
+    }
+    m_writingImpl.reset();
+    m_readingImpl.reset();
+    m_dims.clear();
+    m_xml = CiftiXML();
+    m_writingFile = "";
+    m_onDiskVersion = CiftiVersion();//for completeness, it gets reset on open anyway
+    m_endianPref = NATIVE;//reset things to defaults
+    setWritingDataTypeNoScaling();//default argument is float32
+}
+
 void CiftiFile::convertToInMemory()
 {
     if (isInMemory()) return;
@@ -231,9 +273,14 @@ void CiftiFile::getColumn(float* dataOut, const int64_t& index) const
 
 void CiftiFile::setCiftiXML(const CiftiXML& xml, const bool useOldMetadata)
 {
+    if (xml.getNumberOfDimensions() == 0) throw CiftiException("setCiftiXML called with 0-dimensional CiftiXML");
+    vector<int64_t> xmlDims = xml.getDimensions();
+    for (size_t i = 0; i < xmlDims.size(); ++i)
+    {
+        if (xmlDims[i] < 1) throw CiftiException("cifti xml dimensions must be greater than zero");
+    }
     m_readingImpl.reset();//drop old implementation, as it is now invalid due to XML (and therefore matrix size) change
     m_writingImpl.reset();
-    if (xml.getNumberOfDimensions() == 0) throw CiftiException("setCiftiXML called with 0-dimensional CiftiXML");
     if (useOldMetadata)
     {
         MetaData newmd = m_xml.getFileMetaData();//make a copy
@@ -242,11 +289,7 @@ void CiftiFile::setCiftiXML(const CiftiXML& xml, const bool useOldMetadata)
     } else {
         m_xml = xml;
     }
-    m_dims = m_xml.getDimensions();
-    for (size_t i = 0; i < m_dims.size(); ++i)
-    {
-        if (m_dims[i] < 1) throw CiftiException("cifti xml dimensions must be greater than zero");
-    }
+    m_dims = xmlDims;
 }
 
 void CiftiFile::setRow(const float* dataIn, const vector<int64_t>& indexSelect)
@@ -300,14 +343,15 @@ void CiftiFile::verifyWriteImpl()
             CiftiOnDiskImpl* testImpl = dynamic_cast<CiftiOnDiskImpl*>(m_readingImpl.get());
             if (testImpl != NULL)
             {
-                AString canonicalCurrent = pathToCanonical(testImpl->getFilename());//returns "" if nonexistant, if unlinked while open
+                AString canonicalCurrent = pathToCanonical(testImpl->getFilename());//returns "" if nonexistent, if unlinked while open
                 if (canonicalCurrent != "" && canonicalCurrent == pathToCanonical(m_writingFile))//these were already absolute
                 {
                     convertToInMemory();//save existing data in memory before we clobber file
                 }
             }
         }
-        m_writingImpl = boost::shared_ptr<CiftiOnDiskImpl>(new CiftiOnDiskImpl(m_writingFile, m_xml, m_onDiskVersion, shouldSwap(m_endianPref)));//this constructor makes new file for writing
+        m_writingImpl = boost::shared_ptr<CiftiOnDiskImpl>(new CiftiOnDiskImpl(m_writingFile, m_xml, m_onDiskVersion, shouldSwap(m_endianPref),
+                                                                               m_writingDataType, m_doWriteScaling, m_minScalingVal, m_maxScalingVal));//this constructor makes new file for writing
         if (m_readingImpl != NULL)
         {
             copyImplData(m_readingImpl.get(), m_writingImpl.get(), m_dims);
@@ -412,7 +456,7 @@ CiftiOnDiskImpl::CiftiOnDiskImpl(const AString& filename)
     if (m_xml.getNumberOfDimensions() + 4 != (int)dimCheck.size()) throw CiftiException("XML does not match number of nifti dimensions in file " + filename + "'");
     for (int i = 4; i < (int)dimCheck.size(); ++i)
     {
-        if (m_xml.getDimensionLength(i - 4) < 1)//CiftiXML will only let this happen with cifti-1
+        if (m_xml.getDimensionLength(i - 4) < 0)//CiftiXML will only let this happen with cifti-1
         {
             m_xml.getSeriesMap(i - 4).setLength(dimCheck[i]);//and only in a series map
         } else {
@@ -424,10 +468,126 @@ CiftiOnDiskImpl::CiftiOnDiskImpl(const AString& filename)
     }
 }
 
-CiftiOnDiskImpl::CiftiOnDiskImpl(const AString& filename, const CiftiXML& xml, const CiftiVersion& version, const bool& swapEndian)
+namespace
+{
+    void warnForBadExtension(const AString& filename, const CiftiXML& myXML)
+    {
+        char junk[16];
+        int32_t intent_code = myXML.getIntentInfo(CiftiVersion(), junk);//use default writing version to check file extension, older version is missing some intent codes
+        switch (intent_code)
+        {
+            case 3000://unknown
+                if (!AString_endsWith(filename, ".nii"))
+                {
+                    cerr << "warning: cifti file of nonstandard mapping combination '" << AString_to_std_string(filename) << "' should be saved ending in .<something>.nii" << endl;
+                }
+                if (AString_endsWith(filename, ".dconn.nii") ||
+                    AString_endsWith(filename, ".dtseries.nii") ||
+                    AString_endsWith(filename, ".pconn.nii") ||
+                    AString_endsWith(filename, ".ptseries.nii") ||
+                    AString_endsWith(filename, ".dscalar.nii") ||
+                    AString_endsWith(filename, ".dfan.nii") ||
+                    AString_endsWith(filename, ".fiberTemp.nii") ||
+                    AString_endsWith(filename, ".dlabel.nii") ||
+                    AString_endsWith(filename, ".pscalar.nii") ||
+                    AString_endsWith(filename, ".pdconn.nii") ||
+                    AString_endsWith(filename, ".dpconn.nii") ||
+                    AString_endsWith(filename, ".pconnseries.nii") ||
+                    AString_endsWith(filename, ".pconnscalar.nii"))
+                {
+                    cerr << "warning: cifti file of nonstandard mapping combination '" << AString_to_std_string(filename) << "' should NOT be saved using an already-used cifti extension, "
+                            << "please choose a different, reasonable cifti extension ending in .<something>.nii" << endl;
+                }
+                break;
+            case 3001:
+                if (!AString_endsWith(filename, ".dconn.nii"))
+                {
+                    cerr << "warning: dense by dense cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .dconn.nii" << endl;
+                }
+                break;
+            case 3002:
+                if (!AString_endsWith(filename, ".dtseries.nii"))
+                {
+                    cerr << "warning: series by dense cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .dtseries.nii" << endl;
+                }
+                break;
+            case 3003:
+                if (!AString_endsWith(filename, ".pconn.nii"))
+                {
+                    cerr << "warning: parcels by parcels cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .pconn.nii" << endl;
+                }
+                break;
+            case 3004:
+                if (!AString_endsWith(filename, ".ptseries.nii"))
+                {
+                    cerr << "warning: series by parcels cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .ptseries.nii" << endl;
+                }
+                break;
+            case 3006://3005 unused in practice
+                if (!(AString_endsWith(filename, ".dscalar.nii") || AString_endsWith(filename, ".dfan.nii") || AString_endsWith(filename, ".fiberTEMP.nii")))
+                {//there are additional special extensions in the standard for this mapping combination (specializations of scalar maps)
+                    //also include workbench's fiberTEMP special extension
+                    cerr << "warning: scalars by dense cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .dscalar.nii" << endl;
+                }
+                break;
+            case 3007:
+                if (!AString_endsWith(filename, ".dlabel.nii"))
+                {
+                    cerr << "warning: labels by dense cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .dlabel.nii" << endl;
+                }
+                break;
+            case 3008:
+                if (!AString_endsWith(filename, ".pscalar.nii"))
+                {
+                    cerr << "warning: scalars by parcels cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .pscalar.nii" << endl;
+                }
+                break;
+            case 3009:
+                if (!AString_endsWith(filename, ".pdconn.nii"))
+                {
+                    cerr << "warning: dense by parcels cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .pdconn.nii" << endl;
+                }
+                break;
+            case 3010:
+                if (!AString_endsWith(filename, ".dpconn.nii"))
+                {
+                    cerr << "warning: parcels by dense cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .dpconn.nii" << endl;
+                }
+                break;
+            case 3011:
+                if (!AString_endsWith(filename, ".pconnseries.nii"))
+                {
+                    cerr << "warning: parcels by parcels by series cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .pconnseries.nii" << endl;
+                }
+                break;
+            case 3012:
+                if (!AString_endsWith(filename, ".pconnscalar.nii"))
+                {
+                    cerr << "warning: parcels by parcels by scalar cifti file '" << AString_to_std_string(filename) << "' should be saved ending in .pconnscalar.nii" << endl;
+                }
+                break;
+            default:
+                CiftiAssert(0);
+                throw CiftiException("internal error, tell the developers what you just tried to do");
+        }
+    }
+}
+
+CiftiOnDiskImpl::CiftiOnDiskImpl(const AString& filename, const CiftiXML& xml, const CiftiVersion& version, const bool& swapEndian,
+                                 const int16_t& datatype, const bool& rescale, const double& minval, const double& maxval)
 {//starts writing new file
+    warnForBadExtension(filename, xml);
     NiftiHeader outHeader;
-    outHeader.setDataType(NIFTI_TYPE_FLOAT32);//actually redundant currently, default is float32
+    if (rescale)
+    {
+        outHeader.setDataTypeAndScaleRange(datatype, minval, maxval);
+    } else {
+        outHeader.setDataType(datatype);
+    }
+    if (outHeader.getNumComponents() != 1)
+    {
+        throw CiftiException("cifti cannot be written with multi-component nifti datatypes (i.e., complex, RGB)");
+    }
     char intentName[16];
     int32_t intentCode = xml.getIntentInfo(version, intentName);
     outHeader.setIntent(intentCode, intentName);
@@ -455,6 +615,11 @@ CiftiOnDiskImpl::CiftiOnDiskImpl(const AString& filename, const CiftiXML& xml, c
     m_xml = xml;
 }
 
+void CiftiOnDiskImpl::close()
+{
+    m_nifti.close();//lets this throw when there is a writing problem
+}//don't bother resetting m_xml, this instance is about to be destroyed
+
 void CiftiOnDiskImpl::getRow(float* dataOut, const vector<int64_t>& indexSelect, const bool& tolerateShortRead) const
 {
     m_nifti.readData(dataOut, 5, indexSelect, tolerateShortRead);//5 means 4 reserved (space and time) plus the first cifti dimension
