@@ -40,6 +40,8 @@
     #include <QFile>
 #else
     #include "stdio.h"
+    #include "sys/stat.h"
+    #include "sys/types.h"
     #include "errno.h"
     #define BOOST_FILESYSTEM_VERSION 3
     #include "boost/filesystem.hpp"
@@ -70,6 +72,7 @@ namespace cifti
         void close();
         void seek(const int64_t& position);
         int64_t pos();
+        int64_t size() { return -1; }
         void read(void* dataOut, const int64_t& count, int64_t* numRead);
         void write(const void* dataIn, const int64_t& count);
         ~ZFileImpl();
@@ -88,6 +91,7 @@ namespace cifti
         void close();
         void seek(const int64_t& position);
         int64_t pos();
+        int64_t size() { return m_file.size(); }
         void read(void* dataOut, const int64_t& count, int64_t* numRead);
         void write(const void* dataIn, const int64_t& count);
     };
@@ -104,6 +108,7 @@ namespace cifti
         void close();
         void seek(const int64_t& position);
         int64_t pos();
+        int64_t size();
         void read(void* dataOut, const int64_t& count, int64_t* numRead);
         void write(const void* dataIn, const int64_t& count);
         ~StrFileImpl();
@@ -186,6 +191,12 @@ int64_t BinaryFile::pos()
     return m_impl->pos();
 }
 
+int64_t BinaryFile::size()
+{
+    if (m_curMode == NONE) throw CiftiException("file is not open, can't report size");
+    return m_impl->size();
+}
+
 void BinaryFile::write(const void* dataIn, const int64_t& count)
 {
     CiftiAssert(count >= 0);//not sure about allowing 0
@@ -394,7 +405,7 @@ void QFileImpl::write(const void* dataIn, const int64_t& count)
     while (total < count)
     {
         int64_t maxToWrite = min(count - total, CHUNK_SIZE);
-        writeret = m_file.write((const char*)dataIn, maxToWrite);//QFile probably also chokes on large writes
+        writeret = m_file.write(((const char*)dataIn) + total, maxToWrite);//QFile probably also chokes on large writes
         if (writeret < 1) break;//0 or -1 means error or eof
         total += writeret;
     }
@@ -494,6 +505,14 @@ int64_t StrFileImpl::pos()
     return m_curPos;//we can avoid a call here also
 }
 
+int64_t StrFileImpl::size()
+{
+    struct stat mystat;
+    int result = fstat(fileno(m_file), &mystat);
+    if (result != 0) return -1;
+    return mystat.st_size;
+}
+
 void StrFileImpl::write(const void* dataIn, const int64_t& count)
 {
     if (m_file == NULL) throw CiftiException("write called on unopened StrFileImpl");//shouldn't happen
