@@ -37,6 +37,7 @@
 #include <cstring>
 #include <iostream>
 #include <limits>
+#include "stdint.h"
 
 using namespace std;
 using namespace boost;
@@ -219,7 +220,7 @@ vector<vector<float> > NiftiHeader::getSForm() const
             {
                 for (int j = 0; j < 3; ++j)
                 {
-                    rotmat[i][j] *= m_header.pixdim[i + 1];
+                    rotmat[i][j] *= m_header.pixdim[j + 1];
                 }
             }
             if (m_header.pixdim[0] < 0.0f)//left handed coordinate system, flip the kvec
@@ -239,13 +240,13 @@ vector<vector<float> > NiftiHeader::getSForm() const
             ret[1][3] = m_header.qoffset_y;
             ret[2][3] = m_header.qoffset_z;
         } else {
-            cerr << "found quaternion with length greater than 1 in nifti header" << endl;
+            cerr << "warning: found quaternion with length greater than 1 in nifti header, using ANALYZE coordinates!" << endl;
             ret[0][0] = m_header.pixdim[1];
             ret[1][1] = m_header.pixdim[2];
             ret[2][2] = m_header.pixdim[3];
         }
     } else {//fall back to analyze and complain
-        cerr << "no sform or qform code found, using ANALYZE coordinates!" << endl;
+        cerr << "warning: no sform or qform code found, using ANALYZE coordinates!" << endl;
         ret[0][0] = m_header.pixdim[1];
         ret[1][1] = m_header.pixdim[2];
         ret[2][2] = m_header.pixdim[3];
@@ -265,11 +266,31 @@ vector<vector<float> > NiftiHeader::getSForm() const
         case NIFTI_UNITS_MM:
             break;
         default:
-            cerr << "unrecognized spatial unit in nifti header" << endl;
+            cerr << "warning: unrecognized spatial unit in nifti header" << endl;
     }
     return ret.getMatrix();
 }
 
+double NiftiHeader::getTimeStep() const
+{
+    int timeUnit = XYZT_TO_TIME(m_header.xyzt_units);
+    double ret = m_header.pixdim[4];
+    switch (timeUnit)
+    {
+        case NIFTI_UNITS_USEC:
+            ret /= 1000000.0;
+            break;
+        case NIFTI_UNITS_MSEC:
+            ret /= 1000.0;
+            break;
+        default:
+            cerr << AString_to_std_string("warning: non-time units code " + AString_number(timeUnit) + " used in nifti header, pretending units are seconds") << endl;
+        case NIFTI_UNITS_SEC:
+            break;
+    }
+    return ret;
+}
+
 AString NiftiHeader::toString() const
 {
     AString ret;
@@ -325,6 +346,16 @@ AString NiftiHeader::toString() const
         ret += "qoffset_y: " + AString_number(m_header.qoffset_y) + "\n";
         ret += "qoffset_z: " + AString_number(m_header.qoffset_z) + "\n";
     }
+    ret += "effective sform:\n";
+    vector<vector<float> > tempSform = getSForm();
+    for (int i = 0; i < 3; ++i)
+    {
+        for (int j = 0; j < 4; ++j)
+        {
+            ret += " " + AString_number(tempSform[i][j]);
+        }
+        ret += "\n";
+    }
     ret += "xyzt_units: " + AString_number(m_header.xyzt_units) + "\n";
     ret += "intent_code: " + AString_number(m_header.intent_code) + "\n";
     ret += "intent_name: " + AString_from_latin1(m_header.intent_name, 16) + "\n";
@@ -377,6 +408,13 @@ void NiftiHeader::setIntent(const int32_t& code, const char name[16])
     for (; i < 16; ++i) m_header.intent_name[i] = '\0';
 }
 
+void NiftiHeader::setDescription(const char descrip[80])
+{
+    int i;//custom strncpy-like code to fill nulls to the end
+    for (i = 0; i < 80 && descrip[i] != '\0'; ++i) m_header.descrip[i] = descrip[i];
+    for (; i < 80; ++i) m_header.descrip[i] = '\0';
+}
+
 void NiftiHeader::setSForm(const vector<vector<float> >& sForm)
 {
     CiftiAssert(sForm.size() >= 3);//programmer error to pass badly sized matrix
@@ -386,7 +424,8 @@ void NiftiHeader::setSForm(const vector<vector<float> >& sForm)
         CiftiAssert(sForm[i].size() >= 4);//ditto
         if (sForm[i].size() < 4) throw CiftiException("internal error: setSForm matrix badly sized");
     }
-    m_header.xyzt_units = SPACE_TIME_TO_XYZT(NIFTI_UNITS_MM, NIFTI_UNITS_SEC);//overwrite whatever units we read in
+    int timeUnit = XYZT_TO_TIME(m_header.xyzt_units);
+    m_header.xyzt_units = SPACE_TIME_TO_XYZT(NIFTI_UNITS_MM, timeUnit);//overwrite whatever spatial unit we read in
     for (int i = 0; i < 4; i++)
     {
         m_header.srow_x[i] = sForm[0][i];
@@ -411,9 +450,9 @@ void NiftiHeader::setSForm(const vector<vector<float> >& sForm)
         kvec = -kvec;//because to nifti, "left handed" apparently means "up is down", not "left is right"
     }
     float rotmat[3][3];
-    rotmat[0][0] = ivec[0]; rotmat[1][0] = jvec[0]; rotmat[2][0] = kvec[0];
-    rotmat[0][1] = ivec[1]; rotmat[1][1] = jvec[1]; rotmat[2][1] = kvec[1];
-    rotmat[0][2] = ivec[2]; rotmat[1][2] = jvec[2]; rotmat[2][2] = kvec[2];
+    rotmat[0][0] = ivec[0]; rotmat[1][0] = ivec[1]; rotmat[2][0] = ivec[2];
+    rotmat[0][1] = jvec[0]; rotmat[1][1] = jvec[1]; rotmat[2][1] = jvec[2];
+    rotmat[0][2] = kvec[0]; rotmat[1][2] = kvec[1]; rotmat[2][2] = kvec[2];
     float quat[4];
     if (!MathFunctions::matrixToQuatern(rotmat, quat))
     {
@@ -435,6 +474,13 @@ void NiftiHeader::setSForm(const vector<vector<float> >& sForm)
     }
 }
 
+void NiftiHeader::setTimeStep(const double& seconds)
+{
+    int spaceUnit = XYZT_TO_SPACE(m_header.xyzt_units);//save the current space units so we don't clobber it...
+    m_header.xyzt_units = SPACE_TIME_TO_XYZT(spaceUnit, NIFTI_UNITS_SEC);//overwrite the time part of the units with seconds
+    m_header.pixdim[4] = seconds;
+}
+
 void NiftiHeader::clearDataScaling()
 {
     m_header.scl_slope = 1.0;
@@ -447,6 +493,141 @@ void NiftiHeader::setDataScaling(const double& mult, const double& offset)
     m_header.scl_inter = offset;
 }
 
+namespace
+{
+    template<typename T>
+    struct Scaling
+    {
+        double mult, offset;
+        Scaling(const double& minval, const double& maxval)
+        {
+            std::numeric_limits<T> mylimits;
+            double mymin = mylimits.min();
+            if (!mylimits.is_integer) mymin = -mylimits.max();//again, c++11 can use lowest() instead of these lines
+            mult = (maxval - minval) / ((double)mylimits.max() - mymin);//multiplying is the first step of decoding (after byteswap), so start with the range
+            offset = minval - mymin * mult;//offset is added after multiplying the encoded value by mult
+        }
+    };
+}
+
+void NiftiHeader::setDataTypeAndScaleRange(const int16_t& type, const double& minval, const double& maxval)
+{
+    setDataType(type);
+    switch (type)
+    {
+        case NIFTI_TYPE_RGB24:
+            clearDataScaling();//RGB ignores scaling fields
+            break;
+        case DT_BINARY://currently not supported in read/write functions anyway
+            setDataScaling(maxval - minval, minval);//make the two possible decoded values equal to the min and max
+            break;
+        case NIFTI_TYPE_INT8:
+        {
+            Scaling<int8_t> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_UINT8:
+        {
+            Scaling<uint8_t> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_INT16:
+        {
+            Scaling<int16_t> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_UINT16:
+        {
+            Scaling<uint16_t> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_INT32:
+        {
+            Scaling<int32_t> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_UINT32:
+        {
+            Scaling<uint32_t> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_INT64:
+        {
+            Scaling<int64_t> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_UINT64:
+        {
+            Scaling<uint64_t> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_FLOAT32:
+        case NIFTI_TYPE_COMPLEX64:
+        {
+            Scaling<float> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_FLOAT64:
+        case NIFTI_TYPE_COMPLEX128:
+        {
+            Scaling<double> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        case NIFTI_TYPE_FLOAT128:
+        case NIFTI_TYPE_COMPLEX256:
+        {
+            Scaling<long double> myscale(minval, maxval);
+            setDataScaling(myscale.mult, myscale.offset);
+            break;
+        }
+        default:
+            CiftiAssert(0);
+            throw CiftiException("internal error, report what you did to the developers");
+    }
+}
+
+int NiftiHeader::getNumComponents() const
+{
+    switch (getDataType())
+    {
+        case NIFTI_TYPE_RGB24:
+            return 3;
+            break;
+        case NIFTI_TYPE_COMPLEX64:
+        case NIFTI_TYPE_COMPLEX128:
+        case NIFTI_TYPE_COMPLEX256:
+            return 2;
+            break;
+        case DT_BINARY:
+        case NIFTI_TYPE_INT8:
+        case NIFTI_TYPE_UINT8:
+        case NIFTI_TYPE_INT16:
+        case NIFTI_TYPE_UINT16:
+        case NIFTI_TYPE_INT32:
+        case NIFTI_TYPE_UINT32:
+        case NIFTI_TYPE_FLOAT32:
+        case NIFTI_TYPE_INT64:
+        case NIFTI_TYPE_UINT64:
+        case NIFTI_TYPE_FLOAT64:
+        case NIFTI_TYPE_FLOAT128:
+            return 1;
+            break;
+        default:
+            CiftiAssert(0);
+            throw CiftiException("internal error, report what you did to the developers");
+    }
+}
+
 void NiftiHeader::read(BinaryFile& inFile)
 {
     nifti_1_header buffer1;
@@ -454,6 +635,7 @@ void NiftiHeader::read(BinaryFile& inFile)
     inFile.read(&buffer1, sizeof(nifti_1_header));
     int version = NIFTI2_VERSION(buffer1);
     bool swapped = false;
+    Quirks myquirks;
     try
     {
         if (version == 2)
@@ -465,14 +647,14 @@ void NiftiHeader::read(BinaryFile& inFile)
                 swapped = true;
                 swapHeaderBytes(buffer2);
             }
-            setupFrom(buffer2);
+            myquirks = setupFrom(buffer2, inFile.getFilename());
         } else if (version == 1) {
             if (NIFTI2_NEEDS_SWAP(buffer1))//yes, this works on nifti-1 also
             {
                 swapped = true;
                 swapHeaderBytes(buffer1);
             }
-            setupFrom(buffer1);
+            myquirks = setupFrom(buffer1, inFile.getFilename());
         } else {
             throw CiftiException(inFile.getFilename() + " is not a valid NIfTI file");
         }
@@ -480,58 +662,74 @@ void NiftiHeader::read(BinaryFile& inFile)
         throw CiftiException("error reading NIfTI file " + inFile.getFilename() + ": " + e.whatString());
     }
     m_extensions.clear();
-    char extender[4];
-    inFile.read(extender, 4);
-    int extensions = 0;//if it has extensions in a format we don't know about, don't try to read them
-    if (version == 1 && extender[0] != 0) extensions = 1;//sadly, this is the only thing nifti-1 says about the extender bytes
-    if (version == 2 && extender[0] == 1 && extender[1] == 0 && extender[2] == 0 && extender[3] == 0) extensions = 1;//from http://nifti.nimh.nih.gov/nifti-2 as of 4/4/2014:
-    if (extensions == 1)//"extentions match those of NIfTI-1.1 when the extender bytes are 1 0 0 0"
+    if (myquirks.no_extender)
     {
-        int64_t extStart;
-        if (version == 1)
+        int min_offset = 352;
+        if (version == 2) min_offset = 544;
+        cerr << AString_to_std_string("warning: in file '" + inFile.getFilename() + "', vox_offset is " + AString_number(m_header.vox_offset) +
+        ", nifti standard specifies that it should be at least " + AString_number(min_offset) + ", assuming malformed file with no extender") << endl;
+    } else {
+        char extender[4];
+        inFile.read(extender, 4);
+        int extensions = 0;//if it has extensions in a format we don't know about, don't try to read them
+        if (version == 1 && extender[0] != 0) extensions = 1;//sadly, this is the only thing nifti-1 says about the extender bytes
+        if (version == 2 && extender[0] == 1 && extender[1] == 0 && extender[2] == 0 && extender[3] == 0) extensions = 1;//from http://nifti.nimh.nih.gov/nifti-2 as of 4/4/2014:
+        if (extensions == 1)//"extentions match those of NIfTI-1.1 when the extender bytes are 1 0 0 0"
         {
-            extStart = 352;
-        } else {
-            CiftiAssert(version == 2);
-            extStart = 544;
-        }
-        CiftiAssert(inFile.pos() == extStart);
-        while(extStart + 2 * sizeof(int32_t) <= (size_t)m_header.vox_offset)
-        {
-            int32_t esize, ecode;
-            inFile.read(&esize, sizeof(int32_t));
-            if (swapped) ByteSwapping::swap(esize);
-            inFile.read(&ecode, sizeof(int32_t));
-            if (swapped) ByteSwapping::swap(ecode);
-            if (esize < 8 || esize + extStart > m_header.vox_offset) break;
-            boost::shared_ptr<NiftiExtension> tempExtension(new NiftiExtension());
-            if ((size_t)esize > 2 * sizeof(int32_t))//don't try to read 0 bytes
+            int64_t extStart;
+            if (version == 1)
             {
-                tempExtension->m_bytes.resize(esize - 2 * sizeof(int32_t));
-                inFile.read(tempExtension->m_bytes.data(), esize - 2 * sizeof(int32_t));
+                extStart = 352;
+            } else {
+                CiftiAssert(version == 2);
+                extStart = 544;
+            }
+            CiftiAssert(inFile.pos() == extStart);
+            while(extStart + 2 * sizeof(int32_t) <= (size_t)m_header.vox_offset)
+            {
+                int32_t esize, ecode;
+                inFile.read(&esize, sizeof(int32_t));
+                if (swapped) ByteSwapping::swap(esize);
+                inFile.read(&ecode, sizeof(int32_t));
+                if (swapped) ByteSwapping::swap(ecode);
+                if (esize < 8 || esize + extStart > m_header.vox_offset) break;
+                boost::shared_ptr<NiftiExtension> tempExtension(new NiftiExtension());
+                if ((size_t)esize > 2 * sizeof(int32_t))//don't try to read 0 bytes
+                {
+                    tempExtension->m_bytes.resize(esize - 2 * sizeof(int32_t));
+                    inFile.read(tempExtension->m_bytes.data(), esize - 2 * sizeof(int32_t));
+                }
+                tempExtension->m_ecode = ecode;
+                m_extensions.push_back(tempExtension);
+                extStart += esize;//esize includes the two int32_ts
             }
-            tempExtension->m_ecode = ecode;
-            m_extensions.push_back(tempExtension);
-            extStart += esize;//esize includes the two int32_ts
         }
     }
     m_isSwapped = swapped;//now that we know there were no errors (because they throw), complete the internal state
     m_version = version;
 }
 
-void NiftiHeader::setupFrom(const nifti_1_header& header)
+NiftiHeader::Quirks NiftiHeader::setupFrom(const nifti_1_header& header, const AString& filename)
 {
-    if (header.sizeof_hdr != sizeof(nifti_1_header)) throw CiftiException("incorrect sizeof_hdr");
+    Quirks ret;
+    if (header.sizeof_hdr != sizeof(nifti_1_header)) throw CiftiException("incorrect sizeof_hdr in file '" + filename + "'");
     const char magic[] = "n+1\0";//only support single-file nifti
-    if (strncmp(header.magic, magic, 4) != 0) throw CiftiException("incorrect magic");
-    if (header.dim[0] < 1 || header.dim[0] > 7) throw CiftiException("incorrect dim[0]");
+    if (strncmp(header.magic, magic, 4) != 0) throw CiftiException("incorrect magic in file '" + filename + "'");
+    if (header.dim[0] < 1 || header.dim[0] > 7) throw CiftiException("incorrect dim[0] in file '" + filename + "'");
     for (int i = 0; i < header.dim[0]; ++i)
     {
-        if (header.dim[i + 1] < 1) throw CiftiException("dim[" + AString_number(i + 1) + "] < 1");
+        if (header.dim[i + 1] < 1) throw CiftiException("dim[" + AString_number(i + 1) + "] < 1 in file '" + filename + "'");
+    }
+    if (header.vox_offset < 352)
+    {
+        if (header.vox_offset < 348)
+        {
+            throw CiftiException("file '" + filename + "' has invalid vox_offset: " + AString_number(header.vox_offset));
+        }
+        ret.no_extender = true;
     }
-    if (header.vox_offset < 352) throw CiftiException("incorrect vox_offset");
     int numBits = typeToNumBits(header.datatype);
-    if (header.bitpix != numBits) throw CiftiException("datatype disagrees with bitpix");
+    if (header.bitpix != numBits) cerr << AString_to_std_string("warning: datatype disagrees with bitpix in file '" + filename + "'") << endl;
     m_header.sizeof_hdr = header.sizeof_hdr;//copy in everything, so we don't have to fake anything to print the header as read
     for (int i = 0; i < 4; ++i)//mostly using nifti-2 field order to make it easier to find if things are missed
     {
@@ -574,24 +772,27 @@ void NiftiHeader::setupFrom(const nifti_1_header& header)
     m_header.intent_code = header.intent_code;
     for (int i = 0; i < 16; ++i) m_header.intent_name[i] = header.intent_name[i];
     m_header.dim_info = header.dim_info;
+    return ret;
 }
 
-void NiftiHeader::setupFrom(const nifti_2_header& header)
+NiftiHeader::Quirks NiftiHeader::setupFrom(const nifti_2_header& header, const AString& filename)
 {
-    if (header.sizeof_hdr != sizeof(nifti_2_header)) throw CiftiException("incorrect sizeof_hdr");
+    Quirks ret;
+    if (header.sizeof_hdr != sizeof(nifti_2_header)) throw CiftiException("incorrect sizeof_hdr in file '" + filename + "'");
     const char magic[] = "n+2\0\r\n\032\n";//only support single-file nifti
     for (int i = 0; i < 8; ++i)
     {
-        if (header.magic[i] != magic[i]) throw CiftiException("incorrect magic");
+        if (header.magic[i] != magic[i]) throw CiftiException("incorrect magic in file '" + filename + "'");
     }
-    if (header.dim[0] < 1 || header.dim[0] > 7) throw CiftiException("incorrect dim[0]");
+    if (header.dim[0] < 1 || header.dim[0] > 7) throw CiftiException("incorrect dim[0] in file '" + filename + "'");
     for (int i = 0; i < header.dim[0]; ++i)
     {
-        if (header.dim[i + 1] < 1) throw CiftiException("dim[" + AString_number(i + 1) + "] < 1");
+        if (header.dim[i + 1] < 1) throw CiftiException("dim[" + AString_number(i + 1) + "] < 1 in file '" + filename + "'");
     }
-    if (header.vox_offset < 352) throw CiftiException("incorrect vox_offset");
-    if (header.bitpix != typeToNumBits(header.datatype)) throw CiftiException("datatype disagrees with bitpix");
+    if (header.vox_offset < 544) throw CiftiException("file '" + filename + "' has invalid vox_offset: " + AString_number(header.vox_offset));//haven't noticed any nifti-2 with bad vox_offset yet, and all cifti files have a big extension, so they have to have used it correctly
+    if (header.bitpix != typeToNumBits(header.datatype)) cerr << AString_to_std_string("warning: datatype disagrees with bitpix in file '" + filename + "'") << endl;
     memcpy(&m_header, &header, sizeof(nifti_2_header));
+    return ret;
 }
 
 int NiftiHeader::typeToNumBits(const int64_t& type)
@@ -631,7 +832,7 @@ int NiftiHeader::typeToNumBits(const int64_t& type)
             return 256;
             break;
         default:
-            throw CiftiException("incorrect datatype code");
+            throw CiftiException("incorrect nifti datatype code");
     }
 }
 
@@ -706,6 +907,12 @@ void NiftiHeader::swapHeaderBytes(nifti_2_header& header)
 void NiftiHeader::write(BinaryFile& outFile, const int& version, const bool& swapEndian)
 {
     if (!canWriteVersion(version)) throw CiftiException("unable to write NIfTI version " + AString_number(version) + " for file " + outFile.getFilename());
+    double junk1, junk2;
+    int16_t datatype = getDataType();
+    if (getDataScaling(junk1, junk2) && ((datatype & 0x70) > 0 || datatype >= 1536))
+    {//that hacky expression is to detect 16, 32, 64, 1536, 1792, and 2048
+        cerr << "warning: writing nifti file with scaling factor and floating point datatype" << endl;
+    }
     const char padding[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
     int64_t voxOffset;
     if (version == 2)
